from os import path
import shlex
import ast
import re
import struct

from .lowlevel import ConfigInfo

from . import dispatcher


MAPNAMES = ['sup', 'usr', 'cpu', 'ovl']
BATNAMES = ['ibat0', 'ibat1', 'ibat2', 'ibat3', 'dbat0', 'dbat1', 'dbat2', 'dbat3']


class CodeLine(dict):
    def __getattr__(self, attrname):
        return self[attrname]

    def __setattr__(self, attrname, attrval):
        self[attrname] = attrval


def iter_configinfo_names():
    n = 1
    while True:
        yield 'Configfile-%d' % n
        n += 1


def is_safe(expr):
    chexpr = re.sub(r'\b(0x[0-9a-f]+|0b[01]+|\b[1-9][0-9]*|\b0\b)', '', expr.lower())
    for char in chexpr:
        if char not in '+-|&()':
            return False

    return True


def sub_constants(expr):
    K = dict(PMDT_InvalidAddress=0xA00, PMDT_Available=0xA01)

    for k, v in K.items():
        expr = re.sub(r'\b%s\b' % k, '(%s)' % hex(v), expr)

    return expr


def parse_configinfo(src_path):
    filenames = {}

    linelist = []
    chunks = {'': linelist} # must sort as first

    for line in open(src_path):
        words = shlex.split(line, comments=True, posix=True)
        if len(words) == 0: continue

        if len(words) == 1 and words[0].startswith('[') and words[0].endswith(']'):
            linelist = []
            chunks[words[0][1:-1]] = linelist
            continue

        worddict = CodeLine()
        linelist.append(worddict)
        for word in words:
            k, sep, v = word.partition('=')
            if sep:
                v1, sep, v2 = v.partition('=') # the second = delimits a filename
                if sep and k.endswith('Offset'):
                    worddict[k] = v1
                    filenames[k] = v2
                else:
                    worddict[k] = v

    # do some cleanup: replace all instances of BASE with ROMImageBaseOffset
    base = '-0x30C000' # bad fallback, don't skip ROMImageBaseOffset
    lines = chunks['']
    for words in lines:
        for k, v in words.items():
            if k == 'ROMImageBaseOffset':
                base = v

    base = '(%s)' % base

    for header, lines in chunks.items():
        for words in lines:
            for k, v in list(words.items()):
                if k == 'BootstrapVersion':
                    words[k] = v.encode('mac_roman')
                    continue

                if k == 'brpn':
                    v2 = re.sub(r'\bBASE\b', base, v)
                    is_relative = (v != v2)
                    if is_safe(v2):
                        words[k] = is_relative, eval(v2)
                    continue

                v2 = re.sub(r'\bBASE\b', base, v)
                v2 = sub_constants(v2)
                if is_safe(v2):
                    words[k] = eval(v2)

    return chunks, filenames


def insert_and_assert(binary, insertee, offset):
    new_len = offset + len(insertee)
    binary.extend(b'\0' * (new_len - len(binary)))

    existing = binary[offset:offset+len(insertee)]
    if any(existing): # premature optimisation
        for a, b in zip(insertee, existing):
            if a != 0 and b != 0 and a != b:
                raise ValueError('inserting over something else @%X' % offset)

    binary[offset:offset+len(insertee)] = insertee


def checksum_image(binary, ofs):
    # ugly, but iterating the right was is painfully slow

    byte_lanes = [sum(binary[i::8]) for i in range(8)]

    zeroed_byte_lanes = list(byte_lanes)
    for j in range(ofs, ofs+40):
        zeroed_byte_lanes[j % 8] -= binary[j]

    sum32 = [lane % (1<<32) for lane in zeroed_byte_lanes]

    sum64 = sum(lane << (k * 8) for (k, lane) in enumerate(reversed(zeroed_byte_lanes)))
    sum64 %= 1 << 64

    allsums = b''.join(x.to_bytes(4, byteorder='big') for x in sum32)
    allsums += sum64.to_bytes(8, byteorder='big')

    return allsums


def build(src):
    cilist = []
    for ciname in iter_configinfo_names():
        try:
            cilist.append(parse_configinfo(path.join(src, ciname)))
        except (FileNotFoundError, NotADirectoryError):
            break

    if len(cilist) == 0: raise dispatcher.WrongFormat

    # Expand this as we go
    rom = bytearray()

    # Now we go through every configinfo and insert it (oh hell)
    for ci, filenames in reversed(cilist):
        fields = {key: 0 for key in ConfigInfo._fields}
        lowmem = bytearray()
        pagemap = bytearray()
        segptrs = [bytearray(128) for _ in MAPNAMES]
        batmap = bytearray() # will be padded to 128
        batptrs = [0 for _ in MAPNAMES]

        for header, lines in ci.items():
            if header == '':
                for words in lines:
                    for k, v in words.items():
                        if k in fields:
                            fields[k] = v

                            # The parallel filenames dict tells us what data to put at that address
                            if k in filenames:
                                blob = dispatcher.build(path.join(src, filenames[k]))
                                try:
                                    insert_and_assert(rom, blob, v - fields['ROMImageBaseOffset'])
                                except ValueError:
                                    raise ValueError('Could not insert %r at %s' % (filenames[k], v - fields['ROMImageBaseOffset']))

            elif header == 'LowMemory':
                for words in lines:
                    lowmem.extend(struct.pack('>LL', words.address, words.value))

            elif header == 'PageMappingInfo':
                for words in lines:
                    if 'segment_ptr_here' in words:
                        mapidx = MAPNAMES.index(words.map.lower())
                        struct.pack_into('>LL', segptrs[mapidx], 8 * words.segment_ptr_here, len(pagemap), words.segment_register)

                    elif 'special_pmdt' in words:
                        key = 'PageMap%sOffset' % words.special_pmdt.upper()
                        fields[key] = len(pagemap)

                    elif 'pmdt_page_offset' in words:
                        long2 = words.phys_page << 12 | words.attr
                        pagemap.extend(struct.pack('>HHL', words.pmdt_page_offset, words.pages_minus_1, long2))

            elif header == 'BatMappingInfo':
                for words in lines:
                    if 'bat_ptr_here' in words:
                        batidx = BATNAMES.index(words.bat_ptr_here.lower())
                        mapidx = MAPNAMES.index(words.map.lower())
                        fourbits = len(batmap) // 8
                        shift = 4 * (7 - batidx)
                        batptrs[mapidx] |= fourbits << shift

                    elif 'bepi' in words:
                        ubat = lbat = 0

                        ubat |= words.bepi & 0xFFFE0000 # trailing zeroes hopefully
                        ubat |= words.bl_128k << 2
                        ubat |= words.vs << 1
                        ubat |= words.vp

                        lbat |= words.unk23 << 8
                        lbat |= words.wim << 4
                        lbat |= words.ks << 3
                        lbat |= words.ku << 2
                        lbat |= words.pp

                        is_relative, brpn = words.brpn # special case in parse_configinfo
                        if is_relative: lbat |= 0x200
                        lbat = (lbat + brpn) & 0xFFFFFFFF

                        batmap.extend(struct.pack('>LL', ubat, lbat))

        # Get the awkward array data into the struct
        fields['SegMap32SupInit'] = segptrs[0]
        fields['SegMap32UsrInit'] = segptrs[1]
        fields['SegMap32CPUInit'] = segptrs[2]
        fields['SegMap32OvlInit'] = segptrs[3]
        fields['BatMap32SupInit'] = batptrs[0]
        fields['BatMap32UsrInit'] = batptrs[1]
        fields['BatMap32CPUInit'] = batptrs[2]
        fields['BatMap32OvlInit'] = batptrs[3]
        fields['BATRangeInit'] = batmap

        # Great, now we'll neaten up fields and blat it out
        lowmem.extend(b'\0\0\0\0')

        flat = bytearray(0x1000)
        ptr = len(flat)

        ptr -= len(lowmem)
        insert_and_assert(flat, lowmem, ptr)
        fields['MacLowMemInitOffset'] = ptr

        if len(pagemap) > 0:
            ptr -= len(pagemap)
            insert_and_assert(flat, pagemap, ptr)
            fields['PageMapInitOffset'] = ptr
            fields['PageMapInitSize'] = len(pagemap)

        insert_and_assert(flat, ConfigInfo.pack(**fields), 0)

        # Insert the ConfigInfo struct!
        configinfo_offset = -fields['ROMImageBaseOffset'] # this var used below
        insert_and_assert(rom, flat, configinfo_offset)

        rom.extend(b'\0' * (fields['ROMImageSize'] - len(rom)))

    # let's do a cheeky checksum!
    cksum = checksum_image(rom, configinfo_offset)
    insert_and_assert(rom, cksum, configinfo_offset) # overwrites start of ConfigInfo

    return bytes(rom)



import importlib

import os
from os import path


FORMATS = '''
    bootinfo
    parcels
    powerpc
    supermario
'''.split()


class WrongFormat(Exception):
    pass


def build_dir(p):
    for fmt in FORMATS:
        mod = importlib.import_module('..%s_build' % fmt, __name__)
        try:
            data = mod.build(p)
        except WrongFormat:
            continue

        print(fmt)
        return data

    raise WrongFormat


def build(p):
    parent = path.dirname(path.abspath(p))
    name = path.basename(path.abspath(p))

    # try building the file from a directory
    for np in [p + '.src', p]:
        try:
            data = build_dir(np)
        except WrongFormat:
            pass
        else:
            break
    else:
        # fall back on just reading the file (boring, I know!)
        with open(p, 'rb') as f:
            data = f.read()

    return data


def dump(binary, dest_path, toplevel=False):
    if not toplevel:
        with open(dest_path, 'wb') as f:
            f.write(binary)

        dest_path += '.src'

    for fmt in FORMATS:
        mod = importlib.import_module('..%s_dump' % fmt, __name__)
        try:
            arg = binary
            if isinstance(arg, tuple) and fmt != 'bootinfo': arg = arg[0] # strip found resource fork
            mod.dump(arg, dest_path)
            print(fmt)
            break
        except WrongFormat:
            pass



from collections import defaultdict, Counter
import os
from os import path
from shlex import quote
import struct
import hashlib

from . import dispatcher

from .slow_lzss import decompress
from .lowlevel import PrclNodeStruct, PrclChildStruct
from .pef_info import suggest_name


HEADER_COMMENT = """
# Automated dump of Toolbox Parcels (magic number 'prcl')

#parcel_type [metadata...]
\t#child_type [metadata...] [src=*[.lzss]]
\t\t#null_terminated_strings_instead_of_src_file

# Parcel types are four bytes (child types are unimportant)
#   'prop': match and edit an existing DT node
#   'node': create a new DT node
#   'rom ': Power Macintosh ROM image
#   'psum': black/whitelists for computing DT checksum

#  Flag    Struct   Meaning of known flag
#  -----   ------   --------------------------------------------
#  F0000   parcel   (bitmask) number of new 'special' DT node
#  00200   parcel   edit DT node only if required for boot disk
#  00010   parcel   use only once
#  00008   parcel   match DT node if: ('device_type' == b field)
#  00004   parcel      AND  ('compatible' contains a field
#  00002   parcel           OR   parent 'name' == a field
#  00001   parcel           OR   'name' == a field)
#  -----   ------   --------------------------------------------
#  F0000   child    (bitmask) number of 'special' parent
#  00080   child    create DT prop under 'special' DT node above
#  00100   child    DT prop is for boot debugging only
#  00040   child    delete existing DT prop (vs create)
#  00020   child    do not replace existing DT prop
#  00010   child    use only once
#  00004   child    checksum enabled (crc32)

""".strip()


def quickhash(foo):
    return hashlib.sha512(foo).hexdigest()


def walk_tree(binary):
    """Get low level representation of tree

    e.g. [(prclnodetuple, [prclchildtuple, ...]), ...]
    """

    prclnode = None

    parents = []
    for i in iter(lambda: prclnode.link if prclnode else struct.unpack_from('>12xI', binary)[0], 0):
        prclnode = PrclNodeStruct.unpack_from(binary, offset=i)
        
        children = []
        for j in range(i + PrclNodeStruct.size, i + prclnode.hdr_size, prclnode.child_size):
            prclchild = PrclChildStruct.unpack_from(binary, offset=j)

            children.append(prclchild)

        parents.append((prclnode, children))

    return parents


def unique_binary_tpl(prclchild):
    return (prclchild.ptr, prclchild.packedlen, prclchild.compress)


def guess_binary_name(parent_struct, child_struct, adjacent_name, data):
    # 4 MB ROM-in-RAM image
    if parent_struct.ostype == child_struct.ostype == 'rom ':
        return 'MacROM'

    # Native (PCI) driver with an embedded name and version
    ndrv_name = suggest_name(data)
    if ndrv_name: return ndrv_name

    # A "special" property called by its actual name
    if parent_struct.flags & 0xF0000 or child_struct.flags & 0x80:
        return child_struct.name

    # A driver property with an adjacent name property
    if 'AAPL,MacOS,PowerPC' in child_struct.name and adjacent_name:
        return adjacent_name

    # A lanLib (for netbooting)
    if child_struct.name == 'lanLib,AAPL,MacOS,PowerPC':
        return parent_struct.a + '_lanLib'

    return ''


def dump(binary, dest_dir):
    if not binary.startswith(b'prcl'): raise dispatcher.WrongFormat

    os.makedirs(dest_dir, exist_ok=True)

    basic_structure = walk_tree(binary)

    # Decompress everything
    unpacked_dict = {}
    binary_of = lambda child: unpacked_dict[unique_binary_tpl(child)]
    binary_counts = Counter()
    for prclnode, children in basic_structure:
        for prclchild in children:
            binary_counts[unique_binary_tpl(prclchild)] += 1

            data = binary[prclchild.ptr:prclchild.ptr+prclchild.packedlen]
            if prclchild.compress == 'lzss': data = decompress(data)

            unpacked_dict[unique_binary_tpl(prclchild)] = data

    filename_dict = {} # maps binary data to a filename
    for prclnode, children in basic_structure:
        # A fragment prop may have an adjacent prop giving it a name, get this ready
        adjacent_name = None
        for check_child in children:
            if check_child.name == 'code,AAPL,MacOS,name':
                adjacent_name = unpacked_dict[unique_binary_tpl(check_child)].rstrip(b'\0').decode('ascii')

        # Best guess original-ish name for this binary
        for prclchild in children:
            if prclchild.ostype not in ('cstr', 'csta'):
                base = guess_binary_name(
                    parent_struct=prclnode,
                    child_struct=prclchild,
                    adjacent_name=adjacent_name,
                    data=binary_of(prclchild),
                )
                filename_dict[binary_of(prclchild)] = base

    # Post-process to ensure that all names are unique
    used_names = Counter(filename_dict.values())
    for binary, filename in list(filename_dict.items()):
        if used_names[filename] > 1:
            if filename: filename += '-'
            filename += quickhash(binary)
            filename_dict[binary] = filename 

    filename_dict = {b: (fn+'.pef' if b.startswith(b'Joy!peff') else fn) for (b, fn) in filename_dict.items()}

    # Dump blobs to disk
    for data, filename in filename_dict.items():
        dispatcher.dump(data, path.join(dest_dir, filename))

    # Get printing!!!
    with open(path.join(dest_dir, 'Parcelfile'), 'w') as f:
        f.write(HEADER_COMMENT + '\n\n')

        for prclnode, children in basic_structure:
            line = quote(prclnode.ostype)
            line += ' flags=0x%05x' % prclnode.flags
            if prclnode.a: line += ' a=%s' % quote(prclnode.a)
            if prclnode.b: line += ' b=%s' % quote(prclnode.b)

            print(line, file=f)

            for prclchild in children:
                line = '\t%s' % quote(prclchild.ostype)
                line += ' flags=0x%05x' % prclchild.flags
                if prclchild.name: line += ' name=%s' % quote(prclchild.name)

                if prclchild.ostype not in ('cstr', 'csta'):
                    filename = filename_dict[binary_of(prclchild)]
                    if prclchild.compress == 'lzss': filename += '.lzss'
                    line += ' src=%s' % quote(filename)

                if binary_counts[unique_binary_tpl(prclchild)] > 1:
                    line += ' deduplicate=1'

                if prclnode.ostype == 'psum' and prclchild.ostype == 'csta':
                    if prclchild is children[0]: line += "  # [5] Property whitelist:"
                    if prclchild is children[1]: line += "  # [4] Node 'name' whitelist:"
                    if prclchild is children[2]: line += "  # [3] Node 'name' blacklist:"
                    if prclchild is children[3]: line += "  # [2] Node 'device-type' whitelist:"
                    if prclchild is children[4]: line += "  # [1] Node 'device-type' blacklist:"

                print(line, file=f)

                if prclchild.ostype in ('cstr', 'csta'):
                    strangs = unpacked_dict[unique_binary_tpl(prclchild)].split(b'\0')[:-1]
                    for s in strangs:
                        line = '\t\t%s' % quote(s.decode('ascii'))

                        print(line, file=f)

            print(file=f)



import struct

def tuple_str2bytes(tpl):
    return tuple(x.encode('ascii') if isinstance(x, str) else x for x in tpl)

def tuple_bytes2str(tpl):
    return tuple(x.rstrip(b'\0').decode('ascii') if isinstance(x, bytes) else x for x in tpl)

class StringStruct(struct.Struct):
    """A Struct that works with str instead of bytes"""

    def unpack(self, *args, **kwargs):
        orig = super().unpack(*args, **kwargs)
        return orig.__class__(tuple_bytes2str(orig))

    def unpack_from(self, *args, **kwargs):
        orig = super().unpack_from(*args, **kwargs)
        return orig.__class__(tuple_bytes2str(orig))

    def pack(self, *args, **kwargs):
        return super().pack(*tuple_str2bytes(args), **kwargs)

    def pack_into(self, buf, offset, *args, **kwargs):
        return super().pack_into(buf, offset, *tuple_str2bytes(args), **kwargs)



# Routines to fiddle with 'cfrg' (Code Fragment) metadata resources

# These are required only because the code fragments
# referenced by a cfrg are usually at specific offsets
# in a data fork, rather than in a resource. When we
# manipulate a data fork, any corresponding cfrg must be
# adjusted accordingly. This un-Mac-like scheme originated
# as a way to allow code fragments to be memory-mapped,
# which is impossible in a frequently-repacked resource fork.


import struct


# Internal: where are the fields that must be edited?
def get_dfrkoffset_field_positions(cfrg):
    # old-style cfrg only, seems to work fine...
    entry_cnt, = struct.unpack_from('>L', cfrg, 28)

    ctr = 32

    for i in range(entry_cnt):
        if len(cfrg) < ctr + 43: break

        if cfrg[ctr + 23] == 1: # kDataForkCFragLocator
            yield ctr + 24

        ctr += 42 + 1 + cfrg[ctr + 42]
        while ctr % 4: ctr += 1


# Tell this resource that you moved the PEFs that it references in the data fork
def adjust_dfrkoffset_fields(cfrg, delta):
    cfrg = bytearray(cfrg)

    for field in get_dfrkoffset_field_positions(cfrg):
        ofs, = struct.unpack_from('>L', cfrg, field)
        ofs += delta
        struct.pack_into('>L', cfrg, field, ofs)

    return bytes(cfrg)


# Get the (start, stop) offset range of PEFs in the data fork
def get_dfrk_range(cfrg_list, dfrk_len):
    left = dfrk_len
    right = 0

    for cfrg in cfrg_list:
        for field in get_dfrkoffset_field_positions(cfrg):
            my_left, = struct.unpack_from('>L', cfrg, field)
            left = min(my_left, left)

            my_right, = struct.unpack_from('>L', cfrg, field + 4)
            if my_right == 0:
                right = dfrk_len
            else:
                right = max(right, my_left + my_right)

    return left, right



# Some scrounged code to give name/version suggestions for NDRVs


import struct


MAGIC = b'Joy!peff'


class PEF:
    CONT_HEAD_FMT = '>4s4s4s5I2HI'
    CONT_HEAD_LEN = struct.calcsize(CONT_HEAD_FMT)
    
    SEC_HEAD_FMT = '>i5I4B'
    SEC_HED_LEN = struct.calcsize(SEC_HEAD_FMT)

    def __init__(self, data):
        if not data.startswith(MAGIC): raise ValueError('not a pef')

        (magic, fourcc, arch, ver,
        timestamp, old_def_ver, old_imp_ver, cur_ver,
        sec_count, inst_sec_count, reserv) = struct.unpack_from(self.CONT_HEAD_FMT, data)

        sec_earliest = len(data)
        sec_latest = 0

        self.sections = []
        self.sectypes = []
        self.headeroffsets = []

        self.code = None

        for i in range(sec_count):
            sh_offset = self.CONT_HEAD_LEN + self.SEC_HED_LEN*i

            (sectionName, sectionAddress, execSize,
            initSize, rawSize, containerOffset,
            regionKind, shareKind, alignment, reserved) = struct.unpack_from(self.SEC_HEAD_FMT, data, sh_offset)

            the_sec = data[containerOffset : containerOffset + rawSize]

            if regionKind == 0 and execSize == initSize == rawSize:
                the_sec = bytearray(the_sec)
                self.code = the_sec

            self.sections.append(the_sec)
            self.sectypes.append(regionKind)
            self.headeroffsets.append(sh_offset)

            sec_earliest = min(sec_earliest, containerOffset)
            sec_latest = max(sec_latest, containerOffset + rawSize)

        if any(data[sec_latest:]):
            print('nonzero trailing data from', hex(sec_latest), 'to', hex(len(data)), ' ... will cause incorrect output')

        self.padmult = 1
        while len(data) % (self.padmult * 2) == 0:
            self.padmult *= 2

        self.header = data[:sec_earliest]

    def __bytes__(self):
        accum = bytearray(self.header)

        for i in range(len(self.sections)):
            the_sec = self.sections[i]
            hoff = self.headeroffsets[i]

            while len(accum) % 16:
                accum.append(0)

            new_off = len(accum)
            new_len = len(the_sec)

            accum.extend(the_sec)

            struct.pack_into('>I', accum, hoff + 20, new_off)

            if the_sec is self.code:
                for i in range(8, 20, 4):
                    struct.pack_into('>I', accum, hoff + i, new_len)

        while len(accum) % self.padmult != 0:
            accum.extend(b'\x00')

        return bytes(accum)


def pidata(packed):
    def pullarg(from_iter):
        arg = 0
        for i in range(4):
            cont = next(from_iter)
            arg <<= 7
            arg |= cont & 0x7f
            if not (cont & 0x80): break
        else:
            raise ValueError('arg spread over too many bytes')
        return arg

    packed = iter(packed)
    unpacked = bytearray()

    for b in packed:
        opcode = b >> 5
        arg = b & 0b11111 or pullarg(packed)

        if opcode == 0b000: # zero
            count = arg
            unpacked.extend(b'\0' * count)

        elif opcode == 0b001: # blockCopy
            blockSize = arg
            for i in range(blockSize):
                unpacked.append(next(packed))

        elif opcode == 0b010: # repeatedBlock
            blockSize = arg
            repeatCount = pullarg(packed) + 1
            rawData = bytes(next(packed) for n in range(blockSize))
            for n in range(repeatCount):
                unpacked.extend(rawData)

        elif opcode == 0b011 or opcode == 0b100: # interleaveRepeatBlockWithBlockCopy
            commonSize = arg                     # or interleaveRepeatBlockWithZero
            customSize = pullarg(packed)
            repeatCount = pullarg(packed)

            if opcode == 0b011:
                commonData = bytes(next(packed) for n in range(commonSize))
            else:
                commonData = b'\0' * commonSize

            for i in range(repeatCount):
                unpacked.extend(commonData)
                for j in range(customSize):
                    unpacked.append(next(packed))
            unpacked.extend(commonData)

        else:
            raise ValueError('unknown pidata opcode/arg %s/%d' % (bin(opcode), arg))
            return

    return bytes(unpacked)


def parse_version(num):
    maj, minbug, stage, unreleased = num.to_bytes(4, byteorder='big')

    maj = '%x' % maj
    minor, bugfix = '%02x' % minbug

    if stage == 0x80:
        stage = 'f'
    elif stage == 0x60:
        stage = 'b'
    elif stage == 0x40:
        stage = 'a'
    elif stage == 0x20:
        stage = 'd'
    else:
        stage = '?'

    unreleased = '%d' % unreleased

    vers = maj + '.' + minor

    if bugfix != '0':
        vers += '.' + bugfix

    if (stage, unreleased) != ('f', '0'):
        vers += stage + unreleased

    return vers


def pstring_or_cstring(s):
    plen = s[0]
    pstr = s[1:][:plen]
    cstr = s.rstrip(b'\0')
    if b'\0' in pstr or plen + 1 > len(s):
        return cstr
    else:
        return pstr


def suggest_name(pef):
    if not pef.startswith(b'Joy!peff'): return

    try:
        pef = PEF(pef)

        for sectype, section in zip(pef.sectypes, pef.sections):
            if sectype == 2: section = pidata(section)

            if section and sectype in (1, 2):
                hdr_ofs = section.find(b'mtej')
                if hdr_ofs != -1:
                    sig, strvers, devnam, drvvers = struct.unpack_from('>4s L 32s L', section, hdr_ofs)                

                    # devnam *should* be a 32-byte pascal string, but not if someone forgot the "\p"...
                    devnam = pstring_or_cstring(devnam)

                    sugg = devnam.decode('mac_roman') + '-' + parse_version(drvvers)
                    return sugg
    except:
        pass # do not complain about corrupt PEFs



from os import path
import re
import zlib
import sys
import macresources

try:
    from .fast_lzss import compress
except ImportError:
    from .slow_lzss import compress

from . import dispatcher
from . import cfrg_rsrc


def append_checksum(binary):
    cksum = ('\r\\ h# %08X' % zlib.adler32(binary)).encode('ascii')
    binary.extend(cksum)


# Fix a subtle incompatibility between pre/post-v7.8 scripts & trampolines
# (this is ugly)
def edit_bootscript_for_elf(script, tramp):
    oldprop = b'AAPL,toolbox-image,lzss'
    newprop = b'AAPL,toolbox-parcels'

    matrix = (oldprop in script, newprop in script, oldprop in tramp, newprop in tramp)

    if matrix == (True, False, False, True):
        print('Bootscript older than MacOS.elf (fixing %s => %s)' % (oldprop.decode('ascii'), newprop.decode('ascii')), file=sys.stderr)
        script = script.replace(oldprop, newprop)
    elif matrix == (False, True, True, False):
        print('Bootscript newer than MacOS.elf (fixing %s => %s)' % (newprop.decode('ascii'), oldprop.decode('ascii')), file=sys.stderr)
        script = script.replace(newprop, oldprop)
    else:
        return script

    return script


def build(src):
    try:
        with open(path.join(src, 'Bootscript'), 'rb') as f:
            booter = bytearray(f.read().replace(b'\n', b'\r'))
    except (NotADirectoryError, FileNotFoundError):
        raise dispatcher.WrongFormat

    elf = dispatcher.build(path.join(src, 'MacOS.elf'))
    booter[:] = edit_bootscript_for_elf(booter, elf)

    has_checksum = (b'adler32' in booter)

    constants = dict()
    constant_spans = dict()
    for m in re.finditer(rb'h#\s+([A-Fa-f0-9]+)\s+constant\s+([-\w]+)', booter):
        key = m.group(2).decode('ascii')
        val = int(m.group(1), 16)
        constants[key] = val
        constant_spans[key] = m.span(1)

    booter.append(4) # EOT

    for f in ['elf-offset', 'elf-size']:
        assert f in constants

    if 'elf-offset' in constants:
        # special case: pad according to residual info in script
        booter.extend(b'\0' * (constants['elf-offset'] - len(booter)))

        constants['elf-offset'] = len(booter)
        booter.extend(elf)
        constants['elf-size'] = len(booter) - constants['elf-offset']

    if 'lzss-offset' in constants:
        base = 'lzss'
    else:
        base = 'parcels'

    if base + '-offset' in constants:
        constants[base + '-offset'] = len(booter)
        for attempt in ['MacROM', 'Parcels']:
            try:
                data = dispatcher.build(path.join(src, attempt))
            except:
                pass
            else:
                break
        else:
            raise FileNotFoundError

        if not data.startswith(b'prcl'): data = compress(data)
        booter.extend(data)

        constants[base + '-size'] = len(booter) - constants[base + '-offset']

    constants['info-size'] = len(booter)

    for key, (start, stop) in reversed(sorted(constant_spans.items())):
        insert = ('%X' % constants[key]).zfill(stop - start).encode('ascii')
        assert start + len(insert) == stop
        booter[start:stop] = insert

    if has_checksum: append_checksum(booter)

    # Add a System Enabler (or even just 'vers' information)
    rsrcfork = []
    try:
        datafork = open(path.join(src, 'SysEnabler'), 'rb').read()
        rsrcfork = list(macresources.parse_rez_code(open(path.join(src, 'SysEnabler.rdump'), 'rb').read()))

        while len(booter) % 16: booter.append(0)
        delta = len(booter)
        booter.extend(datafork)
        if len(datafork) > 0 and has_checksum: append_checksum(booter)

        for r in rsrcfork:
            if r.type == b'cfrg':
                r.data = cfrg_rsrc.adjust_dfrkoffset_fields(r.data, delta)

    except FileNotFoundError:
        pass

    return bytes(booter), rsrcfork



from shlex import split
from os import path
import struct
from binascii import crc32

from .lowlevel import PrclNodeStruct, PrclChildStruct, MAGIC

try:
    from .fast_lzss import compress
except ImportError:
    from .slow_lzss import compress

from . import dispatcher


class CodeLine(dict):
    def __getattr__(self, attrname):
        return self[attrname]

    def __setattr__(self, attrname, attrval):
        self[attrname] = attrval


def get_indent_level(from_str):
    if from_str.startswith('\t\t'):
        return 2
    elif from_str.startswith('\t'):
        return 1
    else:
        return 0

def get_keys(from_list, **available):
    ret = CodeLine()

    for k, v in available.items():
        ret[k] = v('')

    for i in from_list:
        k, _, v = i.partition('=')
        fmt = available[k]
        ret[k] = fmt(v)

    return ret

def gethex(from_str):
    if not from_str: return 0
    if from_str.lower().startswith('0x'):
        return int(from_str[2:], base=16)
    else:
        return int(from_str)

def getbool(from_str):
    from_str = from_str.lower()
    if from_str.strip() in ('', 'no', 'n', 'false', 'f', '0'):
        return False
    else:
        return True

class PdslParseError(Exception):
    pass

def build(src):
    if not path.exists(path.join(src, 'Parcelfile')): raise dispatcher.WrongFormat
    node_list = []

    with open(path.join(src, 'Parcelfile')) as f:
        try:
            for line_num, line in enumerate(f, start=1):
                level = get_indent_level(line)
                pieces = split(line, comments=True, posix=True)

                if not pieces: continue

                if level == 0:
                    # parcel node
                    new = get_keys(pieces[1:], flags=gethex, a=str, b=str)
                    new.ostype = pieces[0]
                    new.children = []
                    node_list.append(new)

                elif level == 1:
                    # parcel child
                    new = get_keys(pieces[1:], flags=gethex, name=str, src=str, deduplicate=getbool)
                    new.ostype = pieces[0]
                    new.data = bytearray()
                    new.compress = ''

                    if new.src:
                        if not path.isabs(new.src): # look rel to Parcelfile
                            new.src = path.join(src, new.src)

                        a, b = path.splitext(new.src)
                        if b.lower() == '.lzss':
                            new.src = a
                            new.compress = 'lzss'

                        new.data = dispatcher.build(new.src)
                        new.unpackedlen = len(new.data)
                        if new.compress == 'lzss':
                            new.data = compress(new.data)
                        new.packedlen = len(new.data)

                    node_list[-1].children.append(new)

                elif level == 2:
                    # some C strings to add to the data
                    child = node_list[-1].children[-1]
                    assert not child.src
                    for x in pieces:
                        child.data.extend(x.encode('mac_roman') + b'\0')
                        child.packedlen = child.unpackedlen = len(child.data)

        except:
            raise PdslParseError('Line %d' % line_num)

    # Great! Now that we have this cool data structure, turn it into parcels...
    accum = bytearray()

    accum.extend(MAGIC)
    accum.extend(b'\x00\x00\x00\x14')
    hdr_ptr = len(accum)
    accum.extend(bytes(4))  
    accum.extend(bytes(4))

    dedup_dict = {}
    cksum_history = set() # dedup pointers get a checksum when they shouldn't

    for node in node_list:
        # Link previous member to this one
        struct.pack_into('>I', accum, hdr_ptr, len(accum))

        hdr_ptr = len(accum)
        hdr_size = PrclNodeStruct.size + len(node.children)*PrclChildStruct.size
        accum.extend(b'!' * hdr_size)

        # okay, now start blatting data!
        for child in node.children:
            child.data = bytes(child.data) # no more mutability

            if child.deduplicate and child.data in dedup_dict:
                child.ptr = dedup_dict[child.data]
                continue

            child.ptr = len(accum)

            accum.extend(child.data)

            while len(accum) % 4 != 0:
                accum.append(0x99) # this is the only place we pad

            if child.deduplicate:
                dedup_dict[child.data] = child.ptr

        PrclNodeStruct.pack_into(accum, hdr_ptr,
            link=0, ostype=node.ostype, hdr_size=hdr_size, flags=node.flags,
            n_children=len(node.children), child_size=PrclChildStruct.size,
            a=node.a, b=node.b,
        )

        pack_ptr = hdr_ptr + PrclNodeStruct.size

        for child in node.children:
            if child.flags & 4 or child.ptr in cksum_history:
                data = accum[child.ptr:child.ptr+child.packedlen]
                child.cksum = crc32(data)
                cksum_history.add(child.ptr)
            else:
                child.cksum = 0

            PrclChildStruct.pack_into(accum, pack_ptr, **child)
            pack_ptr += PrclChildStruct.size

    return bytes(accum)



from .stringstruct import StringStruct
from .namedtuplestruct import NamedTupleStruct

class MyParcelStruct(NamedTupleStruct, StringStruct):
    pass

MAGIC = b'prcl\x01\x00\x00\x00'

def pad(offset, align):
    x = offset + align - 1
    x -= x % align
    return x

PrclNodeStruct = MyParcelStruct('>I 4s I I I I 32s 32s', name='PrclNodeStruct',
    fields=['link', 'ostype', 'hdr_size', 'flags', 'n_children', 'child_size', 'a', 'b'])

PrclChildStruct = MyParcelStruct('>4s I 4s I I I I 32s', name='PrclChildStruct',
    fields=['ostype', 'flags', 'compress', 'unpackedlen', 'cksum', 'packedlen', 'ptr', 'name'])

SuperMarioHeader = NamedTupleStruct('>I I B B I I H B B L L L L L L B B 4L L L L L', name='SuperMarioHeader',
    fields=['CheckSum', 'ResetPC', 'MachineNumber', 'ROMVersion',
    'ReStartJMP', 'BadDiskJMP', 'ROMRelease', 'PatchFlags', 'unused1',
    'ForeignOSTbl', 'RomRsrc', 'EjectJMP', 'DispTableOff',
    'CriticalJMP', 'ResetEntryJMP', 'RomLoc', 'unused2', 'CheckSum0',
    'CheckSum1', 'CheckSum2', 'CheckSum3', 'RomSize', 'EraseIconOff',
    'InitSys7ToolboxOff', 'SubVers'])

SuperMarioForeignOS = NamedTupleStruct('>7I', name='SuperMarioForeignOS',
    fields=['InitDispatcher', 'EMT1010', 'BadTrap', 'StartSDeclMgr',
    'InitMemVect', 'SwitchMMU', 'InitRomVectors'])

COMBO_FIELDS = {
    0x40 << 56: 'AppleTalk1',
    0x20 << 56: 'AppleTalk2',
    0x30 << 56: 'AppleTalk2_NetBoot_FPU',
    0x08 << 56: 'AppleTalk2_NetBoot_NoFPU',
    0x10 << 56: 'NetBoot',
    0x78 << 56: 'AllCombos',
}

ResHeader = NamedTupleStruct('>L B B H H 6x', name='ResHeader',
    fields=['offsetToFirst', 'maxValidIndex', 'comboFieldSize',
    'comboVersion', 'headerSize'])

ResEntry = NamedTupleStruct('>Q L L 4s h B 256p', name='ResEntry',
    fields=['combo', 'offsetToNext', 'offsetToData', 'rsrcType',
    'rsrcID', 'rsrcAttr', 'rsrcName'])

def ResEntry_padded_len(entry_tuple):
    return pad(ResEntry.size - 256 + len(entry_tuple.rsrcName), 16)

FakeMMHeader = NamedTupleStruct('>4s L L L', name='FakeMMHeader', 
    fields=['MagicKurt', 'MagicC0A00000', 'dataSizePlus12', 'bogusOff'])

ConfigInfo = NamedTupleStruct('>40x lLL lL lL lL lL lL lL 16sLLL LL LLLLbxxx LLLLL L LLLLLL 128s128s128s128s 128s LLLL L L lL LLL', name='ConfigInfo',
    fields=['ROMImageBaseOffset', 'ROMImageSize', 'ROMImageVersion',
    'Mac68KROMOffset', 'Mac68KROMSize', 'ExceptionTableOffset',
    'ExceptionTableSize', 'HWInitCodeOffset', 'HWInitCodeSize',
    'KernelCodeOffset', 'KernelCodeSize', 'EmulatorCodeOffset',
    'EmulatorCodeSize', 'OpcodeTableOffset', 'OpcodeTableSize',
    'BootstrapVersion', 'BootVersionOffset', 'ECBOffset',
    'IplValueOffset', 'EmulatorEntryOffset', 'KernelTrapTableOffset',
    'TestIntMaskInit', 'ClearIntMaskInit', 'PostIntMaskInit',
    'LA_InterruptCtl', 'InterruptHandlerKind', 'LA_InfoRecord',
    'LA_KernelData', 'LA_EmulatorData', 'LA_DispatchTable',
    'LA_EmulatorCode', 'MacLowMemInitOffset', 'PageAttributeInit',
    'PageMapInitSize', 'PageMapInitOffset', 'PageMapIRPOffset',
    'PageMapKDPOffset', 'PageMapEDPOffset', 'SegMap32SupInit', 'SegMap32UsrInit', 'SegMap32CPUInit', 'SegMap32OvlInit', 'BATRangeInit',
    'BatMap32SupInit', 'BatMap32UsrInit', 'BatMap32CPUInit',
    'BatMap32OvlInit', 'SharedMemoryAddr', 'PA_RelocatedLowMemInit',
    'OpenFWBundleOffset', 'OpenFWBundleSize', 'LA_OpenFirmware',
    'PA_OpenFirmware', 'LA_HardwarePriv'])



from setuptools import setup, Extension

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup_args = dict(
    name='tbxi',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.13',
    author='Elliot Nunn',
    author_email='elliotnunn@fastmail.com',
    description='Tools to compile and inspect Macintosh ROM images',
    url='https://github.com/elliotnunn/tbxi',
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: C',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
    ],
    zip_safe=True,
    packages=['tbxi'],
    install_requires=['macresources'],
    entry_points=dict(console_scripts=['tbxi = tbxi.__main__:main']),
    ext_modules=[Extension('tbxi.fast_lzss', ['speedups/fast_lzss.c'])],
)

# http://charlesleifer.com/blog/misadventures-in-python-packaging-optional-c-extensions/

# Yes, it might be a bit extreme to catch SystemExit to find a compiler error...

try:
    setup(**setup_args)
except (SystemExit, Exception):
    setup_args.pop('ext_modules')
    setup(**setup_args)
